@@ -1,5 +1,7 @@
-import random, traceback
+import random, traceback, socket
 from datetime import datetime
+from binascii import unhexlify
+from flask import make_response
 
 from flask import (
     render_template,
@@ -74,45 +76,6 @@ def inject_debug():
     return dict(debug=app.config["DEBUG"])
 
 
-@app.context_processor
-def inject_tor():
-    if app.config["DEBUG"]:
-        return dict(tor_service_id="", tor_enabled=False)
-    if (
-        request.args.get("action", "") == "stoptor"
-        or request.args.get("action", "") == "starttor"
-    ):
-        if hasattr(current_user, "is_admin") and current_user.is_admin:
-            try:
-                current_hidden_services = (
-                    app.controller.list_ephemeral_hidden_services()
-                )
-            except Exception:
-                current_hidden_services = []
-            if (
-                request.args.get("action", "") == "stoptor"
-                and len(current_hidden_services) != 0
-            ):
-                stop_hidden_services(app)
-            if (
-                request.args.get("action", "") == "starttor"
-                and len(current_hidden_services) == 0
-            ):
-                try:
-                    start_hidden_service(app)
-                except Exception as e:
-                    flash(
-                        "Failed to start Tor hidden service.\
-Make sure you have Tor running with ControlPort configured and try again.\
-Error returned: {}".format(
-                            e
-                        ),
-                        "error",
-                    )
-                    return dict(tor_service_id="", tor_enabled=False)
-    return dict(tor_service_id=app.tor_service_id, tor_enabled=app.tor_enabled)
-
-
 ################ Specter global routes ####################
 @app.route("/")
 @login_required
@@ -160,6 +123,7 @@ def fees(blocks):
 
 @app.route("/get_txout_set_info")
 @login_required
+@app.csrf.exempt
 def txout_set_info():
     res = app.specter.rpc.gettxoutsetinfo()
     return res
@@ -167,6 +131,7 @@ def txout_set_info():
 
 @app.route("/get_scantxoutset_status")
 @login_required
+@app.csrf.exempt
 def get_scantxoutset_status():
     status = app.specter.rpc.scantxoutset("status", [])
     app.specter.info["utxorescan"] = status.get("progress", None) if status else None
@@ -176,3 +141,54 @@ def get_scantxoutset_status():
         "active": app.specter.info["utxorescan"] is not None,
         "progress": app.specter.info["utxorescan"],
     }
+
+
+@app.route("/bitcoin.pdf")
+@login_required
+def get_whitepaper():
+    if app.specter.chain == "main":
+        if not app.specter.info["pruned"]:
+            raw_tx = app.specter.rpc.getrawtransaction(
+                "54e48e5f5c656b26c3bca14a8c95aa583d07ebe84dde3b7dd4a78f4e4186e713",
+                False,
+                "00000000000000ecbbff6bafb7efa2f7df05b227d5c73dca8f2635af32a2e949",
+            )
+            outputs = raw_tx.split("0100000000000000")
+            pdf = ""
+            for output in outputs[1:-2]:
+                cur = 6
+                pdf += output[cur : cur + 130]
+                cur += 132
+                pdf += output[cur : cur + 130]
+                cur += 132
+                pdf += output[cur : cur + 130]
+            pdf += outputs[-2][6:-4]
+        else:
+            outputs_prun = app.specter.rpc.multi(
+                [
+                    (
+                        "gettxout",
+                        "54e48e5f5c656b26c3bca14a8c95aa583d07ebe84dde3b7dd4a78f4e4186e713",
+                        i,
+                    )
+                    for i in range(0, 946)
+                ]
+            )
+            pdf = ""
+            for output in outputs_prun[:-1]:
+                cur = 4
+                pdf += output["result"]["scriptPubKey"]["hex"][cur : cur + 130]
+                cur += 132
+                pdf += output["result"]["scriptPubKey"]["hex"][cur : cur + 130]
+                cur += 132
+                pdf += output["result"]["scriptPubKey"]["hex"][cur : cur + 130]
+            pdf += outputs_prun[-1]["result"]["scriptPubKey"]["hex"][4:-4]
+        res = make_response(unhexlify(pdf[16:-16]))
+        res.headers.set("Content-Disposition", "attachment")
+        res.headers.set("Content-Type", "application/pdf")
+        return res
+    else:
+        return render_template(
+            "500.jinja",
+            error="You need a mainnet node to retrieve the whitepaper. Check your node configurations.",
+        )
