@@ -1,10 +1,10 @@
-import copy, hashlib, json, logging, os
+import copy, hashlib, json, logging, os, re
 import time
 from .device import Device
 from .key import Key
 from .util.merkleblock import is_valid_merkle_proof
 from .helpers import der_to_bytes
-from .util.base58 import decode_base58
+from embit import base58
 from .util.descriptor import Descriptor, sort_descriptor, AddChecksum
 from .util.xpub import get_xpub_fingerprint
 from .util.tx import decoderawtransaction
@@ -19,12 +19,13 @@ from .addresslist import AddressList
 from .txlist import TxList
 
 logger = logging.getLogger()
+LISTTRANSACTIONS_BATCH_SIZE = 1000
 
 
 class Wallet:
     # if the wallet is old we import 300 addresses
     IMPORT_KEYPOOL = 300
-    # a gap of 20 addresses is what many wallets do
+    # a gap of 20 addresses is what many wallets do (not used with descriptor wallets)
     GAP_LIMIT = 20
     # minimal fee rate is slightly above 1 sat/vbyte
     # to avoid rounding errors
@@ -48,6 +49,7 @@ class Wallet:
         devices,
         sigs_required,
         pending_psbts,
+        frozen_utxo,
         fullpath,
         device_manager,
         manager,
@@ -79,6 +81,7 @@ class Wallet:
             raise Exception("A device used by this wallet could not have been found!")
         self.sigs_required = int(sigs_required)
         self.pending_psbts = pending_psbts
+        self.frozen_utxo = frozen_utxo
         self.fullpath = fullpath
         self.manager = manager
         self.rpc = self.manager.rpc.wallet(
@@ -95,15 +98,12 @@ class Wallet:
         self._transactions = TxList(
             txs_path, self.rpc, self._addresses, self.manager.chain
         )
-        if not self._transactions.file_exists:
-            self.fetch_transactions()
 
         if address == "":
             self.getnewaddress()
         if change_address == "":
             self.getnewaddress(change=True)
 
-        self.getdata()
         self.update()
         if old_format_detected or self.last_block != last_block:
             self.save_to_file()
@@ -133,13 +133,36 @@ class Wallet:
         """Load transactions from Bitcoin Core"""
         arr = []
         idx = 0
-        batch = 100
         while True:
-            res = self.rpc.listtransactions("*", batch, batch * idx, True)
+            res = self.rpc.listtransactions(
+                "*",
+                LISTTRANSACTIONS_BATCH_SIZE,
+                LISTTRANSACTIONS_BATCH_SIZE * idx,
+                True,
+            )
+            res = [
+                tx
+                for tx in res
+                if tx["txid"] not in self._transactions
+                or not self._transactions[tx["txid"]].get("address", None)
+                or self._transactions[tx["txid"]].get("blockhash", None)
+                != tx.get("blockhash", None)
+                or (
+                    self._transactions[tx["txid"]].get("blockhash", None)
+                    and not self._transactions[tx["txid"]].get("blockheight", None)
+                )  # Fix for Core v19 with Specter v1
+                or self._transactions[tx["txid"]].get("conflicts", [])
+                != tx.get("walletconflicts", [])
+            ]
+            # TODO: Looks like Core ignore a consolidation (self-transfer) going into the change address (in listtransactions)
+            # This means it'll show unconfirmed for us forever...
             arr.extend(res)
             idx += 1
             # not sure if Core <20 returns last batch or empty array at the end
-            if len(res) < batch or len(arr) < batch * idx:
+            if (
+                len(res) < LISTTRANSACTIONS_BATCH_SIZE
+                or len(arr) < LISTTRANSACTIONS_BATCH_SIZE * idx
+            ):
                 break
         txs = dict.fromkeys([a["txid"] for a in arr])
         txids = list(txs.keys())
@@ -151,18 +174,76 @@ class Wallet:
             if txs.get(txid, None) is not None:
                 continue
             txs[txid] = r["result"]
+        # This is a fix for Bitcoin Core versions < v0.20
+        # These do not return the blockheight as part of the `gettransaction` command
+        # So here we check if this property is lacking and if so
+        # query the current block height and manually calculate it.
+        ##################### Remove from here after dropping Core v0.19 support #####################
+        check_blockheight = False
+        for tx in txs.values():
+            if tx.get("confirmations", 0) > 0 and "blockheight" not in tx:
+                check_blockheight = True
+                break
+        if check_blockheight:
+            current_blockheight = self.rpc.getblockcount()
+            for tx in txs.values():
+                if tx.get("confirmations", 0) > 0:
+                    tx["blockheight"] = current_blockheight - tx["confirmations"] + 1
+        ##################### Remove until here after dropping Core v0.19 support #####################
         self._transactions.add(txs)
+        if self.use_descriptors:
+            while (
+                len(
+                    [
+                        tx
+                        for tx in self._transactions
+                        if self._transactions[tx]["category"] != "send"
+                        and not self._transactions[tx]["address"]
+                    ]
+                )
+                != 0
+            ):
+                addresses = [
+                    dict(
+                        address=self.get_address(
+                            idx, change=False, check_keypool=False
+                        ),
+                        index=idx,
+                        change=False,
+                    )
+                    for idx in range(
+                        self._addresses.max_index(change=False),
+                        self._addresses.max_index(change=False) + self.GAP_LIMIT,
+                    )
+                ]
+                change_addresses = [
+                    dict(
+                        address=self.get_address(idx, change=True, check_keypool=False),
+                        index=idx,
+                        change=True,
+                    )
+                    for idx in range(
+                        self._addresses.max_index(change=True),
+                        self._addresses.max_index(change=True) + self.GAP_LIMIT,
+                    )
+                ]
+                self._addresses.add(addresses, check_rpc=False)
+                self._addresses.add(change_addresses, check_rpc=False)
+                self._transactions.add(txs)
 
     def update(self):
+        self.getdata()
         self.get_balance()
         self.check_addresses()
-        self.get_info()
 
     def check_unused(self):
         """Check current receive address is unused and get new if needed"""
         addr = self.address
-        while self.rpc.getreceivedbyaddress(addr, 0) != 0:
-            addr = self.getnewaddress()
+        try:
+            while self.rpc.getreceivedbyaddress(addr, 0) != 0:
+                addr = self.getnewaddress()
+        except Exception as e:
+            logger.error(f"Failed to check for address reuse: {e}")
 
     def check_addresses(self):
         """Checking the gap limit is still ok"""
@@ -278,6 +359,7 @@ class Wallet:
         change_keypool = wallet_dict.get("change_keypool", 0)
         sigs_required = wallet_dict.get("sigs_required", 1)
         pending_psbts = wallet_dict.get("pending_psbts", {})
+        frozen_utxo = wallet_dict.get("frozen_utxo", [])
         fullpath = wallet_dict.get("fullpath", default_fullpath)
         last_block = wallet_dict.get("last_block", None)
 
@@ -310,6 +392,7 @@ class Wallet:
             devices,
             sigs_required,
             pending_psbts,
+            frozen_utxo,
             fullpath,
             device_manager,
             manager,
@@ -326,13 +409,27 @@ class Wallet:
 
     def check_utxo(self):
         try:
+            locked_utxo = self.rpc.listlockunspent()
+            if locked_utxo:
+                self.rpc.lockunspent(True, locked_utxo)
             utxo = self.rpc.listunspent(0)
+            if locked_utxo:
+                self.rpc.lockunspent(False, locked_utxo)
+                for tx in utxo:
+                    if [
+                        _tx
+                        for _tx in locked_utxo
+                        if _tx["txid"] == tx["txid"] and _tx["vout"] == tx["vout"]
+                    ]:
+                        tx["locked"] = True
             # list only the ones we know (have descriptor for it)
             utxo = [tx for tx in utxo if tx.get("desc", "")]
             for tx in utxo:
                 tx_data = self.gettransaction(tx["txid"], 0)
                 tx["time"] = tx_data["time"]
                 tx["category"] = "send"
+                if "locked" not in tx:
+                    tx["locked"] = False
                 try:
                     # get category from the descriptor - recv or change
                     idx = tx["desc"].split("[")[1].split("]")[0].split("/")[-2]
@@ -340,16 +437,13 @@ class Wallet:
                         tx["category"] = "receive"
                 except:
                     pass
-            self.utxo = utxo
-            self.utxo_labels_list = self.getlabels(
-                utxo["address"] for utxo in self.utxo
-            )
+            self.full_utxo = sorted(utxo, key=lambda utxo: utxo["time"], reverse=True)
         except Exception as e:
             logger.error(f"Failed to load utxos, {e}")
-            self.utxo = []
-            self.utxo_labels_list = {}
+            self.full_utxo = []
 
     def getdata(self):
+        self.fetch_transactions()
         self.check_utxo()
         self.get_info()
         # TODO: Should do the same for the non change address (?)
@@ -371,6 +465,10 @@ class Wallet:
             self.change_index += 1
             self.getnewaddress(change=True)
 
+    @property
+    def utxo(self):
+        return [utxo for utxo in self.full_utxo if not utxo["locked"]]
+
     @property
     def json(self):
         return self.to_json()
@@ -398,6 +496,7 @@ class Wallet:
             o["labels"] = self.export_labels()
         else:
             o["pending_psbts"] = self.pending_psbts
+            o["frozen_utxo"] = self.frozen_utxo
             o["last_block"] = self.last_block
         return o
 
@@ -411,6 +510,12 @@ class Wallet:
         delete_file(self._addresses.path)
         delete_file(self._transactions.path)
 
+    @property
+    def use_descriptors(self):
+        if not hasattr(self, "info") or self.info != {}:
+            self.get_info()
+        return "descriptors" in self.info and self.info["descriptors"] == True
+
     @property
     def is_multisig(self):
         return len(self.keys) > 1
@@ -437,6 +542,34 @@ class Wallet:
             del self.pending_psbts[txid]
             self.save_to_file()
 
+    def toggle_freeze_utxo(self, utxo_list):
+        # utxo = ["txid:vout", "txid:vout"]
+        for utxo in utxo_list:
+            if utxo in self.frozen_utxo:
+                try:
+                    self.rpc.lockunspent(
+                        True,
+                        [{"txid": utxo.split(":")[0], "vout": int(utxo.split(":")[1])}],
+                    )
+                except Exception as e:
+                    # UTXO was spent
+                    print(e)
+                    pass
+                self.frozen_utxo.remove(utxo)
+            else:
+                try:
+                    self.rpc.lockunspent(
+                        False,
+                        [{"txid": utxo.split(":")[0], "vout": int(utxo.split(":")[1])}],
+                    )
+                except Exception as e:
+                    # UTXO was spent
+                    print(e)
+                    pass
+                self.frozen_utxo.append(utxo)
+
+        self.save_to_file()
+
     def update_pending_psbt(self, psbt, txid, raw):
         if txid in self.pending_psbts:
             self.pending_psbts[txid]["base64"] = psbt
@@ -465,116 +598,132 @@ class Wallet:
             )
         self.save_to_file()
 
-    def txlist(self, page, limit=100, validate_merkle_proofs=False):
+    def txlist(
+        self,
+        fetch_transactions=True,
+        validate_merkle_proofs=False,
+        current_blockheight=None,
+    ):
+        """Returns a list of all transactions in the wallet's CSV cache - processed with information to display in the UI in the transactions list
+        #Parameters:
+        #    fetch_transactions (bool): Update the TxList CSV caching by fetching transactions from the Bitcoin RPC
+        #    validate_merkle_proofs (bool): Return transactions with validated_blockhash
+        #    current_blockheight (int): Current blockheight for calculating confirmations number (None will fetch the block count from the RPC)
+        """
+        if fetch_transactions or (
+            self.use_descriptors
+            and len(
+                [
+                    tx
+                    for tx in self._transactions
+                    if self._transactions[tx]["category"] != "send"
+                    and not self._transactions[tx]["address"]
+                ]
+            )
+            != 0
+        ):
+            self.fetch_transactions()
         try:
-            rpc_txs = self.rpc.listtransactions(
-                "*", limit + 2, limit * page, True
-            )  # get batch + 2 to make sure you have information about send
-            rpc_txs = [tx for tx in rpc_txs if tx.get("confirmations", 0) >= 0]
-            rpc_txs = [
+            _transactions = [
+                tx.__dict__().copy()
+                for tx in self._transactions.values()
+                if tx["ismine"]
+            ]
+            transactions = sorted(
+                _transactions, key=lambda tx: tx["time"], reverse=True
+            )
+            transactions = [
                 tx
-                for tx in rpc_txs
+                for tx in transactions
                 if (
-                    not tx["walletconflicts"]
+                    not tx["conflicts"]
                     or max(
                         [
                             self.gettransaction(conflicting_tx, 0)["time"]
-                            for conflicting_tx in tx["walletconflicts"]
+                            for conflicting_tx in tx["conflicts"]
                         ]
                     )
-                    < tx["timereceived"]
+                    < tx["time"]
                 )
             ]
-            rpc_txs.reverse()
-            transactions = rpc_txs[:limit]
-        except:
-            return []
-        result = []
-        blocks = {}
-        for tx in transactions:
-            if "confirmations" not in tx:
-                tx["confirmations"] = 0
-
-            # skip change outputs
-            addr = tx.get("address", "")
-            if addr in self._addresses:
-                if self._addresses[addr].change:
-                    continue
-
-            if tx["confirmations"] == 0 and (
-                tx["category"] == "send" and tx["bip125-replaceable"] == "yes"
-            ):
-                raw_tx = decoderawtransaction(
-                    self.gettransaction(tx["txid"], 0)["hex"], self.manager.chain
-                )
-                tx["vsize"] = raw_tx["vsize"]
-
-            # TODO: validate for unique txids only
-            tx["validated_blockhash"] = ""  # default is assume unvalidated
-            if (
-                validate_merkle_proofs is True
-                and tx["confirmations"] > 0
-                and tx.get("blockhash")
-            ):
-                proof_hex = self.rpc.gettxoutproof([tx["txid"]], tx["blockhash"])
-                logger.debug(
-                    f"Attempting merkle proof validation of tx { tx['txid'] } in block { tx['blockhash'] }"
-                )
-                if is_valid_merkle_proof(
-                    proof_hex=proof_hex,
-                    target_tx_hex=tx["txid"],
-                    target_block_hash_hex=tx["blockhash"],
-                    target_merkle_root_hex=None,
-                ):
-                    # NOTE: this does NOT guarantee this blockhash is actually in the real Bitcoin blockchain!
-                    # See merkletooltip.html for details
-                    logger.debug(f"Merkle proof of { tx['txid'] } validation success")
-                    tx["validated_blockhash"] = tx["blockhash"]
+            if not current_blockheight:
+                current_blockheight = self.rpc.getblockcount()
+            result = []
+            blocks = {}
+            for tx in transactions:
+                if not tx.get("blockheight", 0):
+                    tx["confirmations"] = 0
                 else:
-                    logger.warning(
-                        f"Attempted merkle proof validation on {tx['txid']} but failed. This is likely a configuration error but perhaps your node is compromised! Details: {proof_hex}"
+                    tx["confirmations"] = current_blockheight - tx["blockheight"] + 1
+
+                # coinbase tx
+                if tx["category"] == "generate":
+                    if tx["confirmations"] <= 100:
+                        category = "immature"
+
+                if (
+                    tx.get("confirmations") == 0
+                    and tx.get("bip125-replaceable", "no") == "yes"
+                ):
+                    tx["fee"] = self.rpc.gettransaction(tx["txid"]).get("fee", 1)
+
+                if isinstance(tx["address"], str):
+                    tx["label"] = self.getlabel(tx["address"])
+                elif isinstance(tx["address"], list):
+                    tx["label"] = [self.getlabel(address) for address in tx["address"]]
+                else:
+                    tx["label"] = None
+
+                # TODO: validate for unique txids only
+                tx["validated_blockhash"] = ""  # default is assume unvalidated
+                if validate_merkle_proofs is True and tx["confirmations"] > 0:
+                    proof_hex = self.rpc.gettxoutproof([tx["txid"]], tx["blockhash"])
+                    logger.debug(
+                        f"Attempting merkle proof validation of tx { tx['txid'] } in block { tx['blockhash'] }"
                     )
-
-            result.append(tx)
-
-        for tx in result:
-            # find minimal from 3 times:
-            maxtime = 10445238000  # TODO: change after 31 dec 2300 lol
-            tx["time"] = min(
-                tx.get("blocktime", maxtime),
-                tx.get("timereceived", maxtime),
-                tx.get("time", maxtime),
-            )
-
-        # fund duplicates
-        for tx in list(result):
-            if tx["category"] == "send":
-                continue
-            duplicates = [
-                dup
-                for dup in result
-                if dup["txid"] == tx["txid"]
-                and dup["vout"] == tx["vout"]
-                and abs(dup["amount"]) == abs(tx["amount"])
-            ]
-            # we have both receive and send
-            if len(duplicates) > 1:
-                for dup in duplicates:
-                    if dup["category"] == "send":
-                        dup["category"] = "selftransfer"
-                        dup["amount"] = abs(dup["amount"])
+                    if is_valid_merkle_proof(
+                        proof_hex=proof_hex,
+                        target_tx_hex=tx["txid"],
+                        target_block_hash_hex=tx["blockhash"],
+                        target_merkle_root_hex=None,
+                    ):
+                        # NOTE: this does NOT guarantee this blockhash is actually in the real Bitcoin blockchain!
+                        # See merkletooltip.html for details
+                        logger.debug(
+                            f"Merkle proof of { tx['txid'] } validation success"
+                        )
+                        tx["validated_blockhash"] = tx["blockhash"]
                     else:
-                        result.remove(dup)
-        return sorted(result, key=lambda tx: tx["confirmations"])
+                        logger.warning(
+                            f"Attempted merkle proof validation on {tx['txid']} but failed. This is likely a configuration error but perhaps your node is compromised! Details: {proof_hex}"
+                        )
 
-    def gettransaction(self, txid, blockheight=None):
+                result.append(tx)
+            return result
+        except Exception as e:
+            logging.error("Exception while processing txlist: {}".format(e))
+            return []
+
+    def gettransaction(self, txid, blockheight=None, decode=False):
         try:
-            return self._transactions.gettransaction(txid, blockheight)
+            tx_data = self._transactions.gettransaction(txid, blockheight)
+            if decode:
+                return decoderawtransaction(tx_data["hex"], self.manager.chain)
+            return tx_data
         except Exception as e:
             logger.warning("Could not get transaction {}, error: {}".format(txid, e))
 
-    def rescanutxo(self, explorer=None):
-        t = threading.Thread(target=self._rescan_utxo_thread, args=(explorer,))
+    def rescanutxo(self, explorer=None, requests_session=None, only_tor=False):
+        delete_file(self._transactions.path)
+        self.fetch_transactions()
+        t = threading.Thread(
+            target=self._rescan_utxo_thread,
+            args=(
+                explorer,
+                requests_session,
+                only_tor,
+            ),
+        )
         t.start()
 
     def export_labels(self):
@@ -593,7 +742,7 @@ class Wallet:
             for address in addresses:
                 self._addresses.set_label(address, label)
 
-    def _rescan_utxo_thread(self, explorer=None):
+    def _rescan_utxo_thread(self, explorer=None, requests_session=None, only_tor=False):
         # rescan utxo is pretty fast,
         # so we can check large range of addresses
         # and adjust keypool accordingly
@@ -668,13 +817,6 @@ class Wallet:
         # handle missing transactions now
         # if Tor is running, requests will be sent over Tor
         if explorer is not None:
-            try:
-                requests_session = requests.Session()
-                requests_session.proxies["http"] = "socks5h://localhost:9050"
-                requests_session.proxies["https"] = "socks5h://localhost:9050"
-                requests_session.get(explorer)
-            except Exception:
-                requests_session = requests.Session()
             # make sure there is no trailing /
             explorer = explorer.rstrip("/")
             try:
@@ -699,6 +841,31 @@ class Wallet:
                 )
             except Exception as e:
                 logger.warning(f"Failed to fetch data from block explorer: {e}")
+                # retry if using requests_session failed
+                if not only_tor:
+                    try:
+                        # get raw transactions
+                        raws = [
+                            requests.get(f"{explorer}/api/tx/{tx['txid']}/hex").text
+                            for tx in missing
+                        ]
+                        # get proofs
+                        proofs = [
+                            requests.get(
+                                f"{explorer}/api/tx/{tx['txid']}/merkleblock-proof"
+                            ).text
+                            for tx in missing
+                        ]
+                        # import funds
+                        self.rpc.multi(
+                            [
+                                ("importprunedfunds", raws[i], proofs[i])
+                                for i in range(len(raws))
+                            ]
+                        )
+                    except:
+                        logger.warning(f"Failed to fetch data from block explorer: {e}")
+        self.fetch_transactions()
         self.check_addresses()
 
     @property
@@ -713,25 +880,22 @@ class Wallet:
 
     @property
     def blockheight(self):
-        txs = self.rpc.listtransactions("*", 100, 0, True)
-        i = 0
-        while len(txs) == 100:
-            i += 1
-            next_txs = self.rpc.listtransactions("*", 100, i * 100, True)
-            if len(next_txs) > 0:
-                txs = next_txs
-            else:
-                break
-        try:
-            if len(txs) > 0 and "blockheight" in txs[0]:
-                blockheight = (
-                    txs[0]["blockheight"] - 101
-                )  # To ensure coinbase transactions are indexed properly
+        self.fetch_transactions()
+        MAX_BLOCKHEIGHT = 999999999999  # Replace before we reach this height
+        first_tx = sorted(
+            self._transactions.values(),
+            key=lambda tx: tx.get("blockheight", None)
+            if tx.get("blockheight", None)
+            else MAX_BLOCKHEIGHT,
+        )
+        first_tx_blockheight = (
+            first_tx[0].get("blockheight", None) if first_tx else None
+        )
+        if first_tx:
+            if first_tx_blockheight and first_tx_blockheight - 101 > 0:
                 return (
-                    0 if blockheight < 0 else blockheight
-                )  # To ensure regtest don't have negative blockheight
-        except:
-            pass
+                    first_tx_blockheight - 101
+                )  # Give tiny margin to catch edge case of mined coins
         return 481824 if self.manager.chain == "main" else 0
 
     @property
@@ -815,6 +979,10 @@ class Wallet:
             if self.devices[0].device_type in electrum_devices:
                 return {
                     "keystore": {
+                        "ckcc_xfp": int(
+                            "".join(list(reversed(re.findall("..?", key.fingerprint)))),
+                            16,
+                        ),
                         "ckcc_xpub": key.xpub,
                         "derivation": key.derivation.replace("h", "'"),
                         "root_fingerprint": key.fingerprint,
@@ -842,9 +1010,16 @@ class Wallet:
 
         to_return = {"wallet_type": "{}of{}".format(self.sigs_required, len(self.keys))}
         for cnt, device in enumerate(self.devices):
-            key = [key for key in device.keys if key in self.keys][0]
+            keys_matched = [key for key in device.keys if key in self.keys]
+            if keys_matched:
+                key = keys_matched[0]
+            else:
+                return {"error": "Missing key couldn't be found in any device"}
             if device.device_type in electrum_devices:
                 to_return["x{}/".format(cnt + 1)] = {
+                    "ckcc_xfp": int(
+                        "".join(list(reversed(re.findall("..?", key.fingerprint)))), 16
+                    ),
                     "ckcc_xpub": key.xpub,
                     "derivation": key.derivation.replace("h", "'"),
                     "root_fingerprint": key.fingerprint,
@@ -867,7 +1042,11 @@ class Wallet:
 
     def get_balance(self):
         try:
-            balance = self.rpc.getbalances()["watchonly"]
+            balance = (
+                self.rpc.getbalances()["mine"]
+                if self.use_descriptors
+                else self.rpc.getbalances()["watchonly"]
+            )
             # calculate available balance
             locked_utxo = self.rpc.listlockunspent()
             available = {}
@@ -881,6 +1060,7 @@ class Wallet:
                 else:
                     available["trusted"] -= delta
                     available["trusted"] = round(available["trusted"], 8)
+            available["untrusted_pending"] = round(available["untrusted_pending"], 8)
             balance["available"] = available
         except:
             balance = {
@@ -888,24 +1068,35 @@ class Wallet:
                 "untrusted_pending": 0,
                 "available": {"trusted": 0, "untrusted_pending": 0},
             }
-            available["untrusted_pending"] = round(available["untrusted_pending"], 8)
         self.balance = balance
         return self.balance
 
     def keypoolrefill(self, start, end=None, change=False):
+        # Descriptor wallets were introduced in v0.21.0, but upgraded nodes may
+        # still have legacy wallets. Use getwalletinfo to check the wallet type.
+        # The "keypool" for desciptor wallets is automatically refilled
+        if self.use_descriptors and start > 0:
+            return
+
         if end is None:
+            # end is ignored for descriptor wallets
             end = start + self.GAP_LIMIT
+
         desc = self.recv_descriptor if not change else self.change_descriptor
         args = [
             {
                 "desc": desc,
                 "internal": change,
-                "range": [start, end],
                 "timestamp": "now",
-                "keypool": True,
                 "watchonly": True,
             }
         ]
+        if self.use_descriptors:
+            args[0]["active"] = True
+        else:
+            args[0]["keypool"] = True
+            args[0]["range"] = [start, end]
+
         addresses = [
             dict(
                 address=self.get_address(idx, change=change, check_keypool=False),
@@ -917,36 +1108,42 @@ class Wallet:
         self._addresses.add(addresses, check_rpc=False)
 
         if not self.is_multisig:
-            r = self.rpc.importmulti(args, {"rescan": False})
+            if self.use_descriptors:
+                r = self.rpc.importdescriptors(args)
+            else:
+                r = self.rpc.importmulti(args, {"rescan": False})
         # bip67 requires sorted public keys for multisig addresses
         else:
-            # try if sortedmulti is supported
-            r = self.rpc.importmulti(args, {"rescan": False})
-            # doesn't raise, but instead returns "success": False
-            if not r[0]["success"]:
-                # first import normal multi
-                # remove checksum
-                desc = desc.split("#")[0]
-                # switch to multi
-                desc = desc.replace("sortedmulti", "multi")
-                # add checksum
-                desc = AddChecksum(desc)
-                # update descriptor
-                args[0]["desc"] = desc
+            if self.use_descriptors:
+                self.rpc.importdescriptors(args)
+            else:
+                # try if sortedmulti is supported
                 r = self.rpc.importmulti(args, {"rescan": False})
-                # make a batch of single addresses to import
-                arg = args[0]
-                # remove range key
-                arg.pop("range")
-                batch = []
-                for i in range(start, end):
-                    sorted_desc = sort_descriptor(desc, index=i)
-                    # create fresh object
-                    obj = {}
-                    obj.update(arg)
-                    obj.update({"desc": sorted_desc})
-                    batch.append(obj)
-                r = self.rpc.importmulti(batch, {"rescan": False})
+                # doesn't raise, but instead returns "success": False
+                if not r[0]["success"]:
+                    # first import normal multi
+                    # remove checksum
+                    desc = desc.split("#")[0]
+                    # switch to multi
+                    desc = desc.replace("sortedmulti", "multi")
+                    # add checksum
+                    desc = AddChecksum(desc)
+                    # update descriptor
+                    args[0]["desc"] = desc
+                    r = self.rpc.importmulti(args, {"rescan": False})
+                    # make a batch of single addresses to import
+                    arg = args[0]
+                    # remove range key
+                    arg.pop("range")
+                    batch = []
+                    for i in range(start, end):
+                        sorted_desc = sort_descriptor(desc, index=i)
+                        # create fresh object
+                        obj = {}
+                        obj.update(arg)
+                        obj.update({"desc": sorted_desc})
+                        batch.append(obj)
+                    r = self.rpc.importmulti(batch, {"rescan": False})
         if change:
             self.change_keypool = end
         else:
@@ -954,70 +1151,6 @@ class Wallet:
         self.save_to_file()
         return end
 
-    def utxo_on_address(self, address):
-        utxo = [tx for tx in self.utxo if tx["address"] == address]
-        return len(utxo)
-
-    def balance_on_address(self, address):
-        balancelist = [
-            utxo["amount"] for utxo in self.utxo if utxo["address"] == address
-        ]
-        return round(sum(balancelist), 8)
-
-    def utxo_on_label(self, label):
-        return len(
-            [tx for tx in self.utxo if self.utxo_labels_list[tx["address"]] == label]
-        )
-
-    def balance_on_label(self, label):
-        return round(
-            sum(
-                utxo["amount"]
-                for utxo in self.utxo
-                if self.utxo_labels_list[utxo["address"]] == label
-            ),
-            8,
-        )
-
-    def addresses_on_label(self, label):
-        return list(
-            dict.fromkeys(
-                [
-                    address
-                    for address in self.wallet_addresses
-                    if self.getlabel(address) == label
-                ]
-            )
-        )
-
-    def utxo_addresses(self, idx=0, wallet_utxo_batch=100):
-        return list(
-            dict.fromkeys(
-                list(
-                    reversed(
-                        [
-                            utxo["address"]
-                            for utxo in sorted(self.utxo, key=lambda utxo: utxo["time"])
-                        ]
-                    )
-                )[(wallet_utxo_batch * idx) : (wallet_utxo_batch * (idx + 1))]
-            )
-        )
-
-    def utxo_labels(self, idx=0, wallet_utxo_batch=100):
-        return list(
-            dict.fromkeys(
-                list(
-                    reversed(
-                        [
-                            self.utxo_labels_list[utxo["address"]]
-                            for utxo in sorted(self.utxo, key=lambda utxo: utxo["time"])
-                        ]
-                    )
-                )[(wallet_utxo_batch * idx) : (wallet_utxo_batch * (idx + 1))]
-            )
-        )
-
     def setlabel(self, address, label):
         self._addresses.set_label(address, label)
 
@@ -1093,15 +1226,7 @@ class Wallet:
                     "The wallet does not have sufficient funds to make the transaction."
                 )
 
-            if self.available_balance["trusted"] <= sum(amounts):
-                txlist = self.rpc.listunspent(0, 0)
-                b = sum(amounts) - self.available_balance["trusted"]
-                for tx in txlist:
-                    extra_inputs.append({"txid": tx["txid"], "vout": tx["vout"]})
-                    b -= tx["amount"]
-                    if b < 0:
-                        break
-            elif selected_coins != []:
+            if selected_coins != []:
                 still_needed = sum(amounts)
                 for coin in selected_coins:
                     coin_txid = coin.split(",")[0]
@@ -1115,6 +1240,14 @@ class Wallet:
                     raise SpecterError(
                         "Selected coins does not cover Full amount! Please select more coins!"
                     )
+            elif self.available_balance["trusted"] <= sum(amounts):
+                txlist = self.rpc.listunspent(0, 0)
+                b = sum(amounts) - self.available_balance["trusted"]
+                for tx in txlist:
+                    extra_inputs.append({"txid": tx["txid"], "vout": tx["vout"]})
+                    b -= tx["amount"]
+                    if b < 0:
+                        break
 
             # subtract fee from amount of this output:
             # currently only one address is supported, so either
@@ -1128,6 +1261,9 @@ class Wallet:
                 "replaceable": rbf,
             }
 
+            if self.manager.bitcoin_core_version_raw >= 210000:
+                options["add_inputs"] = selected_coins == []
+
             if fee_rate > 0:
                 # bitcoin core needs us to convert sat/B to BTC/kB
                 options["feeRate"] = round((fee_rate * 1000) / 1e8, 8)
@@ -1263,7 +1399,7 @@ class Wallet:
             # for multisig add xpub fields
             if len(self.keys) > 1:
                 for k in self.keys:
-                    key = b"\x01" + decode_base58(k.xpub)
+                    key = b"\x01" + base58.decode_check(k.xpub)
                     if k.fingerprint != "":
                         fingerprint = bytes.fromhex(k.fingerprint)
                     else:
@@ -1357,3 +1493,40 @@ class Wallet:
             return 75 + 34
         # pubkey, signature, 4* P2SH: 16 00 14 20-byte-hash
         return 75 + 34 + 23 * 4
+
+    def addresses_info(self, is_change):
+        """Create a list of (receive or change) addresses from cache and retrieve the
+        related UTXO and amount.
+        Parameters: is_change: if true, return the change addresses else the receive ones.
+        """
+
+        addresses_info = []
+
+        addresses_cache = [
+            v for _, v in self._addresses.items() if v.change == is_change
+        ]
+
+        for addr in addresses_cache:
+
+            addr_utxo = 0
+            addr_amount = 0
+
+            for utxo in [
+                utxo for utxo in self.full_utxo if utxo["address"] == addr.address
+            ]:
+                addr_amount = addr_amount + utxo["amount"]
+                addr_utxo = addr_utxo + 1
+
+            addresses_info.append(
+                {
+                    "index": addr.index,
+                    "address": addr.address,
+                    "label": addr.label,
+                    "amount": addr_amount,
+                    "used": bool(addr.used),
+                    "utxo": addr_utxo,
+                    "type": "change" if is_change else "receive",
+                }
+            )
+
+        return addresses_info
